from abc import ABC, abstractmethod

import numpy as np


def getloss(name):
    if name == "Rmse":
        return Rmse()
    if name == "Mae":
        return Mae()
    if name == "TrimmedRmse":
        return TrimmedRmse()
    if name == "AUC":
        return AUC()
    if name == "Accuracy":
        return Accuracy()
    if name == "MeanPrediction":
        return MeanPrediction()
    if name == "MeanActual":
        return MeanActual()
    else:
        raise NotImplementedError


class Loss(ABC):
    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError

    @property
    def greater_is_better(self):
        return False

    @property
    def convert_to_np_arrays_first(self):
        return False

    def metric(self, pred, y):
        if self.convert_to_np_arrays_first:
            return self._metric(
                np.array(pred, dtype=pred.dtype), np.array(y, dtype=y.dtype)
            )
        else:
            return self._metric(pred, y)

    @abstractmethod
    def _metric(self, pred, y):
        raise NotImplementedError


class Rmse(Loss):
    """Implements RMSE loss"""

    name = "Root Mean Squared Error"
    convert_to_np_arrays_first = True

    def _metric(self, pred, y):
        return np.sqrt(np.mean((y - pred) ** 2))


class Mae(Loss):
    name = "Mean Absolute Error"
    convert_to_np_arrays_first = True

    def _metric(self, pred, y):
        return np.mean(np.abs(y - pred))


class TrimmedRmse(Loss):
    name = "Trimmed RMSE"
    convert_to_np_arrays_first = True

    def _metric(self, pred, y):
        index = trim_outliers(y, return_index=True)
        trmse = np.sqrt(np.mean((y[index] - pred[index]) ** 2))
        return trmse


class MeanPrediction(Loss):
    name = "Mean Prediction"

    def _metric(self, pred, y):
        return np.mean(pred)


class MeanActual(Loss):
    name = "Mean Actual"

    def _metric(self, pred, y):
        return np.mean(y)


def count_false(a):
    """Count number of False (zero) occurences in array"""
    return np.size(a) - np.count_nonzero(a)


def count_true(a):
    """Count number of True (non-zero) occurences in array"""
    return np.count_nonzero(a)


class AUC(Loss):
    name = "AUC"
    greater_is_better = True
    convert_to_np_arrays_first = True

    def _metric(self, pred, y):
        if len(np.unique(y)) != 2:
            raise ValueError("y must consist of two classes (True/False)")
        cutoffs = np.unique(np.append(pred, [-np.inf, np.inf]))
        cutoffs = cutoffs[::-1]  # reverse order
        y = np.array(y)
        pred = np.array(pred)
        tpr = [np.sum((pred > x) & y) / count_true(y) for x in cutoffs]
        fpr = [np.sum((pred > x) & ~y) / count_false(y) for x in cutoffs]
        auc = np.trapz(y=tpr, x=fpr)
        return auc


class Accuracy(Loss):
    name = "Accuracy"
    greater_is_better = True
    threshold = 0.5
    convert_to_np_arrays_first = True

    def _metric(self, pred, y):
        tpr = np.mean((pred > self.threshold) & y)  # true positive rate
        tnr = np.mean((pred < self.threshold) & ~y)  # true negative rate
        accuracy = tpr + tnr
        return accuracy


def trim_outliers(a, lb=0.005, ub=0.995, return_index=False):
    """Drop values at the edges of a distribution"""
    a = np.array(a)
    lb = np.quantile(a, q=lb)
    ub = np.quantile(a, q=ub)
    index = (a >= lb) & (a <= ub)
    if return_index:
        return index
    elif not return_index:
        return a[index]
    else:
        raise ValueError("return_index must only accepts True/False")



from typing import Callable

import altair as alt
import numpy as np
import pandas as pd
from altair.vegalite.v4.schema.channels import Color
from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from profiling.utils import format_time, vega_sanitize


def condexp(
    func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame, y: pd.Series, var: str
):
    """Plot conditional expectations"""
    if is_categorical_dtype(x[var]):
        return condexp_cat(func, x, y, var)
    elif is_bool_dtype(x[var]):
        return condexp_cat(func, x, y, var)
    elif is_object_dtype(x[var]):
        return condexp_cat(func, x, y, var)
    elif is_numeric_dtype(x[var]):
        return condexp_num(func, x, y, var)
    elif is_datetime64_any_dtype(x[var]):
        return condexp_datetime(func, x, y, var)
    else:
        pass


def flatten_index(df):
    """Flatten hiearchical column index"""
    df.columns = df.columns.to_flat_index()
    return df


def condexp_num(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    y: pd.Series,
    var: str,
    n_cut: int = 10,
) -> alt.Chart:
    """Conditional expectations plot for numerical variables"""

    non_missing = ~x[var].isna()
    x = x[non_missing]
    y = y[non_missing]
    pred = func(x).to_numpy()

    df = x.assign(pred=pred, y=y)

    if df[var].nunique() > n_cut:
        groupvar = df[var] + 1e-6 * np.random.uniform(size=len(df))
        df["groups"] = pd.qcut(groupvar, n_cut, duplicates="drop")
    else:
        df["groups"] = df[var]

    df_act = (
        df.groupby("groups")
        .agg({var: "mean", "y": ["mean", "std", "count"]})
        .pipe(flatten_index)
        .assign(__taktile_profiling_value_type__="Actual Value")
        .rename(
            columns={
                ("y", "mean"): y.name,
                ("y", "std"): "y_std",
                ("y", "count"): "y_count",
                (var, "mean"): var,
            }
        )
    )

    df_pred = (
        df.groupby("groups")
        .agg({var: "mean", "pred": "mean"})
        .assign(__taktile_profiling_value_type__="Predicted Value")
        .rename(columns={"pred": y.name})
    )

    df = df_pred.append(df_act, ignore_index=True)

    df_error = (
        df_act.assign(
            lower_bound=lambda x: x[y.name]
            - (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            upper_bound=lambda x: x[y.name]
            + (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            __taktile_profiling_value_type__=lambda x: "Actual Value",
        )
        .rename(columns={y.name: "y.name"})  # to avoid error with melt
        .melt(
            id_vars=[var, "__taktile_profiling_value_type__"],
            value_vars=["lower_bound", "upper_bound"],
            var_name="__taktile_actual_value_ci__",
            value_name=y.name,
        )
    )

    # Create plot
    var_sanitized = vega_sanitize(var, datatype="Q")

    chart = (
        alt.Chart(df)
        .mark_line(point=True)
        .encode(
            x=alt.X(var_sanitized, scale=alt.Scale(zero=False)),
            y=alt.Y(y.name, scale=alt.Scale(zero=False)),
            color=alt.Color(
                "__taktile_profiling_value_type__:N", legend=alt.Legend(title="Type")
            ),
            tooltip=[
                alt.Tooltip("__taktile_profiling_value_type__:N", title="Type"),
                alt.Tooltip(var_sanitized, format=".2f"),
                alt.Tooltip(f"{y.name}:Q", format=".2f"),
            ],
        )
        .interactive()
    )

    chart_error = (
        alt.Chart(df_error)
        .mark_errorband()
        .encode(
            x=alt.X(var_sanitized, scale=alt.Scale(zero=False)),
            y=alt.Y(y.name, scale=alt.Scale(zero=False)),
            color=alt.Color("__taktile_profiling_value_type__:N"),
        )
        .interactive()
    )

    return chart + chart_error


def condexp_cat(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    y: pd.Series,
    var: str,
    n_categories: int = 25,
) -> alt.Chart:
    """Conditional expectations plot for categorical variables"""
    pred = func(x)

    df_agg = (
        x.assign(pred=pred, y=y)
        .astype({var: "category"})
        .groupby(var, as_index=False, observed=True)
        .agg({"y": ["mean", "std", "count"], "pred": ["mean", "count"]})
        .pipe(flatten_index)
        .sort_values(("pred", "count"), ascending=False)
        .head(n_categories)  # keep most frequent
        .sort_values(("pred", "mean"), ascending=False)
        .rename(
            columns={
                (var, ""): var,
                ("y", "mean"): "Actual Value",
                ("pred", "mean"): "Predicted Value",
                ("pred", "count"): "N",
                ("y", "std"): "y_std",
                ("y", "count"): "y_count",
            }
        )
    )

    df_agg[var] = df_agg[var].cat.remove_unused_categories()
    ordered_categories = list(df_agg[var].cat.categories)

    df_pred = df_agg.rename(columns={"Predicted Value": y.name}).assign(
        __taktile_profiling_value_type__="Predicted Value"
    )
    df_act = df_agg.rename(columns={"Actual Value": y.name}).assign(
        __taktile_profiling_value_type__="Actual Value"
    )

    df_error = (
        df_act.assign(
            lower_bound=lambda x: x[y.name]
            - (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            upper_bound=lambda x: x[y.name]
            + (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            __taktile_profiling_value_type__=lambda x: "Actual Value",
        )
        .rename(columns={y.name: "y.name"})  # to avoid error with melt
        .melt(
            id_vars=[var, "__taktile_profiling_value_type__"],
            value_vars=["lower_bound", "upper_bound"],
            var_name="__taktile_actual_value_ci__",
            value_name=y.name,
        )
    )

    # Create plot
    var_sanitized = vega_sanitize(var, datatype="N")

    chart_pred = (
        alt.Chart(df_pred)
        .mark_circle(opacity=0.9)
        .encode(
            x=alt.X(y.name, scale=alt.Scale(zero=False)),
            y=alt.Y(
                var_sanitized,
                sort=ordered_categories,
                scale=alt.Scale(zero=False),
            ),
            color=alt.Color(
                "__taktile_profiling_value_type__:N", legend=alt.Legend(title="Type")
            ),
            tooltip=[
                alt.Tooltip("__taktile_profiling_value_type__:N", title="Type"),
                alt.Tooltip(var_sanitized),
                alt.Tooltip(f"{y.name}:Q", format=".2f"),
            ],
        )
        .interactive()
    )

    chart_act = (
        alt.Chart(df_act)
        .mark_circle(opacity=0.9)
        .encode(
            x=alt.X(y.name, scale=alt.Scale(zero=False)),
            y=alt.Y(var_sanitized, sort=ordered_categories),
            color=alt.Color(
                "__taktile_profiling_value_type__:N",
                legend=alt.Legend(title="Type"),
            ),
            tooltip=[
                alt.Tooltip("__taktile_profiling_value_type__:N", title="Type"),
                alt.Tooltip(var_sanitized),
                alt.Tooltip(f"{y.name}:Q", format=".2f"),
            ],
        )
        .interactive()
    )

    chart_error = (
        alt.Chart(df_error)
        .mark_errorbar(ticks=True)
        .encode(
            x=alt.X(y.name, scale=alt.Scale(zero=False)),
            y=alt.Y(
                var_sanitized, sort=ordered_categories, scale=alt.Scale(zero=False)
            ),
            color=alt.Color("__taktile_profiling_value_type__:N"),
        )
        .interactive()
    )
    chart = chart_pred + chart_act + chart_error

    return chart


def condexp_datetime(func, X, y, var, n_max=100):
    """Conditional expectations plot for datetime variables"""
    non_missing = ~X[var].isna()
    X = X[non_missing]
    y = y[non_missing]
    pred = func(X).to_numpy()

    df = X.assign(pred=pred, y=y)

    df_agg = (
        df.groupby(var, as_index=False, observed=True)
        .agg({"y": ["mean", "std", "count"], "pred": ["mean", "count"]})
        .pipe(flatten_index)
        .sort_values(("pred", "count"), ascending=False)
        .head(n_max)  # keep most frequent
        .rename(
            columns={
                (var, ""): var,
                ("y", "mean"): "Actual Value",
                ("pred", "mean"): "Predicted Value",
                ("pred", "count"): "N",
                ("y", "std"): "y_std",
                ("y", "count"): "y_count",
            }
        )
    )

    df_pred = df_agg.rename(columns={"Predicted Value": y.name}).assign(
        __taktile_profiling_value_type__="Predicted Value"
    )
    df_act = df_agg.rename(columns={"Actual Value": y.name}).assign(
        __taktile_profiling_value_type__="Actual Value"
    )

    df = df_pred.append(df_act, ignore_index=True)

    df_error = (
        df_act.assign(
            lower_bound=lambda x: x[y.name]
            - (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            upper_bound=lambda x: x[y.name]
            + (1.96 * x["y_std"] / np.sqrt(x["y_count"])),
            __taktile_profiling_value_type__=lambda x: "Actual Value",
        )
        .rename(columns={y.name: "y.name"})  # to avoid error with melt
        .melt(
            id_vars=[var, "__taktile_profiling_value_type__"],
            value_vars=["lower_bound", "upper_bound"],
            var_name="__taktile_actual_value_ci__",
            value_name=y.name,
        )
    )

    # Create plot
    var_sanitized = vega_sanitize(var, datatype="T")

    # Format datetimes correctly
    formatted_time = format_time(X, var)

    chart = (
        alt.Chart(df)
        .mark_line(point=True)
        .encode(
            x=var_sanitized,
            y=alt.Y(y.name, scale=alt.Scale(zero=False)),
            color=alt.Color(
                "__taktile_profiling_value_type__:N", legend=alt.Legend(title="Type")
            ),
            tooltip=[
                alt.Tooltip("__taktile_profiling_value_type__:N", title="Type"),
                alt.Tooltip(var_sanitized, format=formatted_time),
                alt.Tooltip(f"{y.name}:Q", format=".2f"),
            ],
        )
        .interactive()
    )
    chart_error = (
        alt.Chart(df_error)
        .mark_errorband()
        .encode(
            x=alt.X(var_sanitized, scale=alt.Scale(zero=False)),
            y=alt.Y(y.name, scale=alt.Scale(zero=False)),
            color=alt.Color("__taktile_profiling_value_type__:N"),
        )
        .interactive()
    )

    return chart + chart_error



import math
from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import lru_cache, partial
from itertools import combinations
from typing import Callable
from typing import Iterable as IterableType
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from profiling.utils import (
    batched_non_object_col_mode,
    batched_numeric_col_mean,
    drop_nas_from_batch,
)


@lru_cache(maxsize=256)
def n_choose_k(n, k):
    """Calculate binomial coefficient

    Numerical stable calculation with caching. Equivalent to:

    factorial(n) / factorial(k) / factorial(n - k)

    which has very large numerators and denominators.
    """
    coef = 1
    k = min(k, n - k)
    for x in range(1, k + 1):
        coef = coef * (n - k + x) // x
    return coef


comb = np.vectorize(
    n_choose_k,
    otypes=[int],
    doc="Vectorized version of 'n choose k', mimics scipy.special.comb",
)


class Explainer(ABC):
    @abstractmethod
    def explain(self, x):
        raise NotImplementedError


def update_columns(base: pd.DataFrame, update: pd.DataFrame) -> pd.DataFrame:
    """Update a base dataframe column-wise with a replacement dataframe

    Parameters
    ----------
    base : pd.DataFrame
        Base dataframe that will be updated
    update : pd.DataFrame
        Values to use when updating

    Returns
    -------
    pd.DataFrame
        Updated DataFrame where base-columns have been replaced with update-columns
    """
    base_columns = base.columns
    base = base.drop(columns=update.columns)
    if (len(base) == 1) & (len(update) > 1):  # numpy-style broadcasting
        base = pd.concat([base] * len(update), axis=0, ignore_index=True)
    base.index = update.index.copy()
    combined = pd.concat([base, update], axis=1)
    return combined.reindex(columns=base_columns)


def reduce_func(func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame):
    """Reduce a function to accept an arbitrary subset of X as input

    Parameters
    ----------
    func : function
        Original function, must accept X as input
    x : pd.DataFrame
        Original dataframe, used for filling in missing columns

    Returns
    -------
    function
        Reduced function that takes an arbitrary subset of X's columns as input
    """

    def func_subset(x_subset, x_, original_func):
        X_updated = update_columns(base=x_, update=x_subset)
        return original_func(X_updated)

    reduced_func = partial(func_subset, x_=x, original_func=func)
    return reduced_func


class ShapExplainer(Explainer):
    def __init__(
        self,
        func: Callable[[pd.DataFrame], pd.Series],
        x: Union[pd.DataFrame, IterableType[pd.DataFrame]],
        y: IterableType[pd.Series] = None,
        profile_columns: Optional[List[str]] = None,
    ):
        """Create a Shap Explainer

        Parameters
        ----------
        func : function
            Function to be explained. Must map from X to array-like
        x : pd.DataFrame
            Background dataset for Shapley value calculation
        profile_columns : list, optional
            Calculate Shapley values only for a subset of columns, by default all columns will be included
        """

        if profile_columns is not None and isinstance(x, pd.DataFrame):
            func = reduce_func(func, x=self._create_background_data(x, y=y))
            x = x[profile_columns]

        self.func = func
        self.x = self._create_background_data(x, y=y)
        self.n_col = len(self.x.columns)

    def _create_background_data(self, x, y=None) -> pd.DataFrame:
        """Create single-row background dataset"""
        if isinstance(x, pd.DataFrame):
            return self._create_background_data_from_full_frame(x)
        elif isinstance(x, Iterable):
            first = drop_nas_from_batch(next(x), next(y))
            return self._created_batched_background_data(first=first, x=x, y=y)

    @staticmethod
    def _created_batched_background_data(
        first: pd.DataFrame, x: IterableType[pd.DataFrame], y: IterableType[pd.Series]
    ):
        background = {}
        for col in first.columns:
            if pd.api.types.is_numeric_dtype(first[col]):
                background[col] = [
                    batched_numeric_col_mean(x=x, y=y, first=first, col=col)
                ]
            else:
                background[col] = [
                    batched_non_object_col_mode(x=x, y=y, first=first, col=col)
                ]
        x_background = pd.DataFrame(background, columns=first.columns).astype(
            first.dtypes
        )
        return x_background

    @staticmethod
    def _create_background_data_from_full_frame(x: pd.DataFrame):
        background = {}
        for col in x.columns:
            if pd.api.types.is_numeric_dtype(x[col]) and not pd.api.types.is_bool_dtype(
                x[col]
            ):
                background[col] = [x[col].mean()]
            else:
                mode = x[col].mode(dropna=True).values
                background[col] = [mode[0]] if mode.size > 0 else [None]
        x_background = pd.DataFrame(background, columns=x.columns).astype(x.dtypes)
        return x_background

    def explain(self, x) -> Tuple[pd.Series, pd.DataFrame]:
        """Calculate Shapley explanations for a dataframe"""
        baselines = []
        explanations = []
        for i in range(len(x)):  # loop over rows
            row = x.iloc[[i]]
            reference, explanation = self._explain_row(row)
            baselines.append(reference)
            explanations.append(explanation)
        baselines = pd.Series(baselines, name="Baseline")
        explanations = pd.DataFrame(explanations, columns=x.columns)
        return baselines, explanations

    def _explain_row(self, row):
        """Calculate Shapley explanations for a single-row dataframe

        We are following Lundberg & Lee (2017) for the Kernel SHAP methodology:

        - Sample possible coalitions (i.e. active/inactive features)
        - Map those coalitions to features and predict
        - Regress predictions on coalition indicators using the SHAP Kernel

        For the final regression, the goal is to minimize the objective

        ||y - phi0 - indicators @ phi||

        with weights w (see Theorem 2 in the paper).

        The SHAP Kernel weights are infinite for the empty and the full coalition.
        This creates two constraints on Shapley values phi0, phi:

        [I]  phi0 == y_empty
        [II] phi0 + np.sum(phi) == y_full

        We minimize the objective on the mixed coalitions (which have finite weights) subject to
        constraints [I] and [II]. This is done by reformulating the objective function.
        First, substitute out phi0 = y_empty using [I]. Second, substitute out the last element
        of phi, phi_k, using [II]. This yields an unconstrained problem that can be solved by
        weighted least squares using numpy only (see code for details).
        """
        empty_coalition = {tuple([0] * self.n_col)}
        full_coalition = {tuple([1] * self.n_col)}

        x_empty = self._map_features(empty_coalition, row)
        x_full = self._map_features(full_coalition, row)

        y_empty = float(self.func(x_empty))
        y_full = float(self.func(x_full))

        if self.n_col == 1:
            # We have two constraints and two unknowns
            phi0 = y_empty
            phi = y_full - phi0
            return phi0, phi

        else:
            # We actually need to sample coalitions and solve the WLS problem
            coalitions = self._sample_coalitions(self.n_col)
            x = self._map_features(coalitions, row)
            y = np.array(self.func(x))

            weights = np.diag(self._coalition_weights(coalitions))
            indicators = np.array(list(coalitions), dtype=int)

            # define phi_ as all thetas but the last and phi_k as the last (k'th) phi
            # define indicators_ and indicators_k as the corresponding coalition indicators
            # we substitute out phi_k using constraint [II], which implies that
            # phi_k == y_full - phi0 - np.sum(phi_)
            phi0 = y_empty  # from [1]
            indicators_ = indicators[:, 0:-1]  # all indicators but the last
            indicators_k = indicators[:, -1]  # last indicator
            lhs = indicators_ - np.expand_dims(indicators_k, 1)
            rhs = (y - phi0) - indicators_k * (y_full - phi0)
            optim = np.linalg.lstsq(a=weights @ lhs, b=weights @ rhs, rcond=None)
            phi_ = optim[0]  # all but last phi
            phi_k = y_full - phi0 - np.sum(phi_)  # recover last phi from constraint
            phi = np.append(phi_, phi_k)  # reconstruct final phi

            return phi0, phi

    def _sample_coalitions(self, n_col):
        """Sample coalition vectors given a maximum coalition size of n_col"""
        budget = 2 * n_col + 2**11  # set sampling budget
        coalitions = set()  # container for collecting coalitions

        for n_out in range(1, math.floor(n_col / 2) + 1):
            # coalitions starting from all columns (but one)
            new_coalitions0 = set(self._unique_coalitions(n_col - n_out, n_out))

            # coalitions starting from no columns (but one)
            if n_out < n_col - n_out:
                new_coalitions1 = set(self._unique_coalitions(n_out, n_col - n_out))
                new_coalitions = new_coalitions0.union(new_coalitions1)
            else:
                new_coalitions = new_coalitions0

            budget -= len(new_coalitions)

            if budget >= 0:
                coalitions.update(new_coalitions)
            else:
                return coalitions

        return coalitions

    def _map_features(self, coalitions, row):
        """Map coalition vectors to original feature space"""
        x_mapped = row.copy().iloc[[0] * len(coalitions)]  # repeat row

        # for each row (coalition) replace values with value from  background dataset
        coalitions_array = np.array(list(coalitions), dtype=bool)

        for i, col in enumerate(x_mapped.columns):
            inside = coalitions_array[:, i]
            x_mapped[col].where(
                cond=inside, other=self.x[col].values[0], inplace=True, try_cast=True
            )
        return x_mapped

    @staticmethod
    def _coalition_weights(coalitions):
        """Calculate the sampling weight of each coalition using the SHAP Kernel"""
        coalitions = np.array(list(coalitions), dtype=bool)
        n_active = np.sum(coalitions, axis=1)
        n_col = coalitions.shape[1]
        weights = (n_col - 1) / (comb(n_col, n_active) * n_active * (n_col - n_active))
        return weights

    @staticmethod
    def _unique_coalitions(n_true, n_false):
        """Calculate all coalitions of n_true "True" and n_false "False" values"""
        length = n_true + n_false
        for indices in combinations(range(length), n_false):
            coalitions = [True] * length  # base case
            for index in indices:
                coalitions[index] = False
            yield tuple(coalitions)



import numpy as np
import pandas as pd
import pytest

from profiling.metrics import AUC, Accuracy, getloss


@pytest.mark.parametrize(
    "pred,y,expected",
    [
        ([0, 1], [0, 1], 1.0),
        ([0, 0], [0, 1], 0.5),
        ([1, 1], [0, 1], 0.5),
        ([0, 0.0, 0], [0, 1, 1], 0.50),
        ([0, 0.5, 0], [0, 1, 1], 0.75),
        ([0, 0.5, 1], [0, 1, 1], 1.0),
    ],
)
def test_auc(pred, y, expected):
    pred = np.array(pred)
    y = np.array(y)
    loss = AUC()
    auc = loss.metric(pred, y)
    assert np.isclose(auc, expected)


def test_auc_error():
    with pytest.raises(ValueError):
        loss = AUC()
        pred = np.array([0, 0])
        y = np.array([1, 1])
        loss.metric(pred, y)


@pytest.mark.parametrize(
    "pred,y,expected",
    [
        ([0, 0.6], [0, 1], 1.0),
        ([0, 0.4], [0, 1], 0.5),
        ([1, 0.6], [0, 1], 0.5),
        ([0, 0.0, 0], [0, 1, 1], 1 / 3),
        ([0, 0.6, 0], [0, 1, 1], 2 / 3),
        ([0, 0.6, 1], [0, 1, 1], 1.0),
    ],
)
def test_accuracy(pred, y, expected):
    pred = np.array(pred)
    y = np.array(y)
    loss = Accuracy()
    auc = loss.metric(pred, y)
    assert np.isclose(auc, expected)


@pytest.mark.parametrize("name", ["Mae", "Rmse", "TrimmedRmse"])
def test_index_invariance(name):
    y_array = np.random.sample([10])
    index = np.random.choice(100, len(y_array), replace=False)
    y_series = pd.Series(y_array)
    y_series.index = index

    pred_array = np.random.sample([10])
    pred_series = pd.Series(pred_array)

    loss = getloss(name)
    metric_series = loss.metric(pred_series, y_series)
    metric_array = loss.metric(pred_array, y_array)

    assert metric_series == metric_array



from typing import Callable

import altair as alt
import numpy as np
import pandas as pd
from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from profiling.utils import format_time, vega_sanitize


def partialdep(func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame, var: str):
    """Plot partial dependence"""
    if is_categorical_dtype(x[var]):
        return partialdep_cat(func=func, x=x, var=var)
    elif is_bool_dtype(x[var]):
        return partialdep_cat(func=func, x=x, var=var)
    elif is_object_dtype(x[var]):
        return partialdep_cat(func=func, x=x, var=var)
    elif is_numeric_dtype(x[var]):
        return partialdep_num(func=func, x=x, var=var)
    elif is_datetime64_any_dtype(x[var]):
        return partialdep_datetime(func, x=x, var=var)
    else:
        pass


def flatten_index(df):
    """Flatten hiearchical column index"""
    df.columns = df.columns.to_flat_index()
    return df


def partialdep_num(
    func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame, var: str
) -> alt.Chart:
    """Plot partial dependence for numerical variables"""
    non_missing = ~x[var].isna()
    x = x[non_missing]

    # Downsample large dataframes
    n_obs = len(x)
    n_max = 100
    if n_obs > n_max:
        id_sample = np.random.choice(n_obs, size=n_max, replace=False)
        x = x.iloc[id_sample]

    # Determine evaluation points (deciles)
    n_cut = 10
    if x[var].nunique() >= n_cut:
        quantiles = np.quantile(
            x[var],
            q=np.linspace(1 / (2 * n_cut), 1 - (1 / (2 * n_cut)), n_cut - 1),
            interpolation="nearest",
        )
    else:
        quantiles = np.sort(x[var].unique())

    # Calculate centered ICE plots
    df_ice = pd.DataFrame()
    for i, quantile in enumerate(quantiles):
        x_mod = x.copy()
        x_mod[var] = quantile
        pred_q = func(x_mod)

        if i == 0:
            offset = pred_q - np.mean(pred_q)  # for centering
            if isinstance(offset, pd.Series):
                offset = offset.values

        pred_q_centered = pred_q - offset
        df_ice_q = pd.DataFrame(
            {
                "id": np.arange(len(x_mod)),
                var: quantile,
                "Predicted Value": pred_q_centered,
            }
        )
        df_ice = df_ice.append(df_ice_q, ignore_index=True, sort=False)

    df_mean = df_ice.groupby(by=var, as_index=False).agg({"Predicted Value": "mean"})

    # Create plot
    var_sanitized = vega_sanitize(var, datatype="Q")

    lines = (
        alt.Chart(df_ice)
        .mark_line(strokeOpacity=0.3, color="grey", strokeWidth=2)
        .encode(
            x=var_sanitized,
            y=alt.Y("Predicted Value:Q", scale=alt.Scale(zero=False)),
            detail="id",
        )
    )
    mean = (
        alt.Chart(df_mean)
        .mark_line(strokeWidth=3, point=True)
        .encode(
            x=var_sanitized,
            y=alt.Y("Predicted Value:Q", scale=alt.Scale(zero=False)),
            tooltip=[
                alt.Tooltip(var_sanitized, format=".2f"),
                alt.Tooltip(f"Predicted Value:Q", format=".2f"),
            ],
        )
        .interactive()
    )
    return lines + mean


def partialdep_cat(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    var: str,
    n_categories: int = 25,
    n_sample: int = 50,
) -> alt.Chart:
    """Plot partial dependence for categorical variables"""
    pred = func(x)

    df_agg = (
        x.assign(pred=pred)
        .astype({var: "category"})
        .groupby(var, as_index=False, observed=True)
        .agg({"pred": ["mean", "count"]})
        .pipe(flatten_index)
        .sort_values(("pred", "count"), ascending=False)
        .head(n_categories)  # keep most frequent
        .sort_values(("pred", "mean"), ascending=False)
        .rename(columns={(var, ""): var})
    )

    df_agg[var] = df_agg[var].cat.remove_unused_categories()
    ordered_categories = list(df_agg[var].cat.categories)

    df_plot = pd.DataFrame()

    for q in ordered_categories:
        x_mod = x.copy()
        x_mod[var] = pd.Series([q] * len(x), dtype=x[var].dtype)
        pred_q = func(x_mod)
        rowid = np.arange(len(x_mod))
        df_plot_q = pd.DataFrame(
            {"id": rowid, var: x_mod[var], "Predicted Value": pred_q.values}
        )
        df_plot = df_plot.append(df_plot_q, ignore_index=True, sort=False)

    # Convert to categorical for plotting
    df_plot[var] = (
        df_plot[var]
        .astype("category")
        .cat.add_categories("Missing")
        .fillna("Missing")
        .cat.remove_unused_categories()
    )

    df_plot_summary = (
        df_plot.groupby(var)
        .agg({"Predicted Value": "mean"})
        .sort_values("Predicted Value", ascending=False)
        .reset_index()
    )

    # Sample
    df_plot_sample = df_plot.query(f"id < {n_sample}")

    # Create plot
    var_sanitized = vega_sanitize(var, "N")

    lines = (
        alt.Chart(df_plot_sample)
        .mark_line(strokeOpacity=0.3, color="grey", strokeWidth=2)
        .encode(
            x=alt.X("Predicted Value", scale=alt.Scale(zero=False)),
            y=alt.Y(
                var_sanitized,
                sort=ordered_categories,
                scale=alt.Scale(zero=False),
            ),
            detail="id",
        )
    )
    mean = (
        alt.Chart(df_plot_summary)
        .mark_line(point=True, width=3)
        .encode(
            x="Predicted Value",
            y=alt.Y(
                var_sanitized,
                sort=ordered_categories,
                scale=alt.Scale(zero=False),
            ),
            tooltip=[
                alt.Tooltip(var_sanitized),
                alt.Tooltip(f"Predicted Value:Q", format=".2f"),
            ],
        )
        .interactive()
    )

    return lines + mean


def partialdep_datetime(
    func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame, var: str
) -> alt.Chart:
    """Plot partial dependence for datetime variables"""
    non_missing = ~x[var].isna()
    x = x[non_missing]

    # Downsample large dataframes
    n_obs = len(x)
    n_max = 100
    if n_obs > n_max:
        id_sample = np.random.choice(n_obs, size=n_max, replace=False)
        x = x.iloc[id_sample]

    # Determine evaluation points (deciles)
    n_cut = 10
    if x[var].nunique() >= n_cut:
        quantiles = np.quantile(
            x[var],
            q=np.linspace(1 / (2 * n_cut), 1 - (1 / (2 * n_cut)), n_cut - 1),
            interpolation="nearest",
        )
    else:
        quantiles = np.sort(x[var].unique())

    # Calculate centered ICE plots
    df_ice = pd.DataFrame()
    for i, quantile in enumerate(quantiles):
        x_mod = x.copy()
        x_mod[var] = quantile
        pred_q = func(x_mod)

        if i == 0:
            offset = pred_q - np.mean(pred_q)  # for centering

        pred_q_centered = pred_q - offset
        df_ice_q = pd.DataFrame(
            {
                "id": np.arange(len(x_mod)),
                var: quantile,
                "Predicted Value": pred_q_centered,
            }
        )
        df_ice = df_ice.append(df_ice_q, ignore_index=True, sort=False)

    df_mean = df_ice.groupby(by=var, as_index=False).agg({"Predicted Value": "mean"})

    # Create plot
    var_sanitized = vega_sanitize(var, datatype="T")

    # Format datetimes correctly
    formatted_time = format_time(x, var)

    lines = (
        alt.Chart(df_ice)
        .mark_line(strokeOpacity=0.3, color="grey", strokeWidth=2)
        .encode(
            x=var_sanitized,
            y=alt.Y("Predicted Value:Q", scale=alt.Scale(zero=False)),
            detail="id",
        )
    )
    mean = (
        alt.Chart(df_mean)
        .mark_line(strokeWidth=3, point=True)
        .encode(
            x=var_sanitized,
            y=alt.Y("Predicted Value:Q", scale=alt.Scale(zero=False)),
            tooltip=[
                alt.Tooltip(var_sanitized, format=formatted_time),
                alt.Tooltip(f"Predicted Value:Q", format=".2f"),
            ],
        )
        .interactive()
    )
    return lines + mean



import json
import logging
import os
from pathlib import Path
from typing import Callable, Dict, List, Optional

import pandas
from fastapi.encoders import jsonable_encoder

from .accuracy import calibration, largest_errors, metrics, roc_curve
from .countplot import countplot
from .dependence import partialdep
from .expectations import condexp
from .importance import varimp
from .shap import ShapExplainer
from .sliceloss import slice_loss
from .t import SliceLoss
from .utils import create_description, df_to_dict

BASE_RESULTS_PATH = os.environ.get("BASE_RESULTS_PATH", "results")


def mkdir(endpoint_name: str, spec_path: str = "") -> Path:
    folder = Path(BASE_RESULTS_PATH) / endpoint_name / spec_path
    folder.mkdir(parents=True, exist_ok=True)
    return folder


def create_shapley_schema(
    func: Callable,
    x: pandas.DataFrame,
    endpoint_name: str,
    profile_columns: List[str],
    explain_input_example: List[Dict],
):
    """Create Shapley schema for endpoint"""

    explainer = ShapExplainer(func=func, x=x, profile_columns=profile_columns)
    folder = mkdir(endpoint_name=endpoint_name)

    input_descriptions = [create_description(x[col]) for col in profile_columns]
    logging.info("Shapley - Creating schema")
    file_path = folder / "schema.json"

    example = pandas.DataFrame(explain_input_example).astype(x[profile_columns].dtypes)

    data = create_explainer_schema(
        explainer_schema_example=example,
        explainer=explainer,
        input_descriptions=input_descriptions,
        explain_input_example=explain_input_example,
    )
    with open(file_path, "w") as f:
        json.dump(jsonable_encoder(data), f)


def create_explainer_schema(
    explainer_schema_example: pandas.DataFrame,
    explainer: ShapExplainer,
    input_descriptions: List[Dict],
    explain_input_example: List[Dict],
) -> Dict:
    """Create schema for explainer input"""

    pred = explainer.func(explainer_schema_example)
    baseline, explanation = explainer.explain(explainer_schema_example)

    example = {
        "inputs": explain_input_example,
        "explanations": df_to_dict(explanation),
        "prediction": list(pred),
        "baseline": list(baseline),
    }
    # build schema
    schema = {"input_descriptions": input_descriptions, "example": example}
    return schema


def create_accuracy(
    func: Callable,
    x: pandas.DataFrame,
    y: pandas.Series,
    endpoint_name: str,
    profile_columns: List[str],
    kind: str,
) -> None:
    """Create accuracy profile for endpoint"""

    folder = mkdir(endpoint_name=endpoint_name, spec_path="accuracy")

    logging.info("Accuracy - Creating calibration plot")
    file_path = folder / "calibration.json"
    chart = calibration(func, x, y)
    chart.save(str(file_path), format="json")

    logging.info("Accuracy - Creating metrics table")
    file_path = folder / "metrics.json"
    metrics_table = metrics(func, x, y, kind)
    with open(file_path, "w") as f:
        f.write(metrics_table)

    logging.info("Accuracy - Finding largest errors")
    file_path = folder / "errors.json"
    errors_table = largest_errors(func, x, y, profile_columns)
    with open(file_path, "w") as f:
        f.write(errors_table)

    if kind == "binary":
        logging.info("Accuracy - Creating ROC curve")
        file_path = folder / "roc_curve.json"
        chart = roc_curve(func, x, y)
        chart.save(str(file_path), format="json")


def create_dependence(
    func: Callable,
    x: pandas.DataFrame,
    endpoint_name: str,
    profile_columns: List[str],
) -> None:
    """Create partial dependence graphs for all variables"""

    x = x.reset_index(drop=True)
    folder = mkdir(endpoint_name=endpoint_name, spec_path="anatomy/partialdep")

    for var in profile_columns:
        logging.info(f"Explanations - Creating partial dependence for {var}")
        file_path = folder / f"{var}.json"
        try:
            chart = partialdep(func=func, x=x, var=var)
            chart.save(str(file_path), format="json")
        except AttributeError:
            logging.warning(f"Could not generate valid chart for {var}")


def create_expectations(
    func: Callable,
    x: pandas.DataFrame,
    y: pandas.Series,
    endpoint_name: str,
    profile_columns: List[str],
) -> None:
    """Create conditional expectations graphs for all variables"""

    x, y = x.reset_index(drop=True), y.reset_index(drop=True)
    folder = mkdir(endpoint_name=endpoint_name, spec_path="anatomy/condexp")

    for var in profile_columns:
        logging.info(f"Explanations - Creating conditional expectations for {var}")
        file_path = folder / f"{var}.json"
        try:
            chart = condexp(func, x, y, var)
            chart.save(str(file_path), format="json")
        except AttributeError as e:
            logging.warning(f"Could not generate valid chart for {var}: {repr(e)}")
        except TypeError as e:
            # happens when object columns cannot be serialized to JSON
            # https://github.com/altair-viz/altair/issues/1355
            logging.warning(f"Could not generate valid chart for {var}: {repr(e)}")


def create_countplot(
    x: pandas.DataFrame, endpoint_name: str, profile_columns: List[str]
) -> None:
    """
    Create count plot graphs for all variables in profile_columns,
    creates folder structure and saves the plots to named files.
    """

    folder = mkdir(endpoint_name=endpoint_name, spec_path="anatomy/countplot")

    for var in profile_columns:
        logging.info(f"Explanations - Creating count plot for {var}")
        file_name = var + ".json"
        file_path = Path(folder, file_name)
        chart = countplot(x[var])
        chart.save(str(file_path), format="json")


def create_slice_loss(
    func: Callable,
    x: pandas.DataFrame,
    y: pandas.Series,
    endpoint_name: str,
    profile_columns: List[str],
    kind: str,
) -> None:
    """Create slice loss graphs for all variables"""

    if kind == "binary":
        loss = SliceLoss.ACCURACY.value
    elif kind == "regression":
        loss = SliceLoss.RMSE.value
    else:
        loss = None
    assert loss is not None

    x, y = x.reset_index(drop=True), y.reset_index(drop=True)
    folder = mkdir(endpoint_name=endpoint_name, spec_path="anatomy/sliceloss")

    for var in profile_columns:
        logging.info(f"Explanations - Creating slice loss for {var}")
        file_path = folder / f"{var}.json"
        chart = slice_loss(func, x, y, var, loss)
        chart.save(str(file_path), format="json")


def create_importance(
    func: Callable,
    x: pandas.DataFrame,
    endpoint_name: str,
    profile_columns: List[str],
    kind: str,
) -> None:

    logging.info("Explanations - Calculating variable importance")
    if kind in ["regression", "binary"]:
        chart, varlist = varimp(func, x, profile_columns=profile_columns)
    else:
        raise NotImplementedError("Unknown endpoint kind:" + kind)

    # Save to disk
    folder = mkdir(endpoint_name=endpoint_name, spec_path="anatomy")

    file_path = folder / "varimp.json"
    chart.save(str(file_path), format="json")

    file_path = folder / "varlist.json"
    with open(file_path, "w") as f:
        json.dump(varlist, f)


def create_summary(
    endpoint_name: str,
    input_names: List[str],
    output_names: str,
    kind: str,
    path: Optional[str] = None,
    explain_path: Optional[str] = None,
    version: Optional[str] = None,
    tags: Optional[List[str]] = None,
    position: Optional[int] = None,
) -> None:

    logging.info(f"Summary - creating endpoint summary for endpoint: {endpoint_name}")
    as_json = {
        "name": endpoint_name,
        "kind": kind,
        "inputs": input_names,
        "output": output_names,
        "tags": tags if tags else [],
        "position": position if position else 0,
    }
    if version is not None:
        as_json["version"] = version
    if path is not None:
        as_json["path"] = path
    if explain_path is not None:
        as_json["explain_path"] = explain_path

    # Save to disk
    folder = mkdir(endpoint_name=endpoint_name)
    with open(folder / "summary.json", "w") as summary:
        json.dump(as_json, summary)



import logging
from typing import Callable, List, Tuple

import altair as alt
import numpy as np
import pandas as pd

from profiling.metrics import getloss
from profiling.utils import df_to_json_table


def metrics(
    func: Callable[[pd.DataFrame], pd.Series], x: pd.DataFrame, y: pd.Series, kind: str
) -> str:
    if kind == "regression":
        loss_list = ["Rmse", "Mae", "TrimmedRmse", "MeanPrediction", "MeanActual"]
    elif kind == "binary":
        loss_list = ["Accuracy", "AUC", "MeanPrediction", "MeanActual"]
    else:
        raise NotImplementedError(f"Unknown endpoint kind: {kind}")

    pred = func(x)
    losses = [getloss(ll) for ll in loss_list]
    loss_names = [loss.name for loss in losses]
    loss_values = [loss.metric(pred, y) for loss in losses]
    table = pd.DataFrame({"Metric": loss_names, "Value": loss_values})

    return df_to_json_table(table, date_format="iso", date_unit="s", double_precision=2)


def largest_errors(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    y: pd.Series,
    profile_columns: List[str],
    n: int = 25,
) -> str:
    """Return observations with largest errors in predictions"""
    pred = func(x).to_numpy()
    y = y.to_numpy()  # to avoid issues due to indices

    df = (
        x[profile_columns]
        .assign(y=y, pred=pred)  # to numpy is needed to help time inference
        .astype({"y": float, "pred": float})  # account for boolean output/predictions
        .assign(Delta=lambda k: np.abs(k["y"] - k["pred"]))
    )

    df = df.sort_values("Delta", ascending=False).head(n)
    df = df.rename(columns={"pred": "Predicted Value", "y": "Actual Value"})

    # Move new columns to the beginning
    cols = df.columns.to_list()
    first = ["Predicted Value", "Actual Value", "Delta"]
    last = [c for c in cols if c not in first]
    df = df.reindex(columns=first + last)

    return df_to_json_table(df, date_format="iso", date_unit="s", double_precision=2)


def calibration(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    y: pd.Series,
    bins: int = 10,
) -> alt.Chart:
    """Plot calibration"""
    preds = func(x).to_numpy()
    pred_grp = pd.qcut(preds, q=bins, duplicates="drop")

    df = pd.DataFrame(
        # conversion to numpy is required so booleans are handled correctly
        {"Actual Value": y.to_numpy(), "Predicted Value": preds, "Group": pred_grp}
    )
    df = df.groupby("Group").agg("mean").reset_index().drop(columns=["Group"])

    # Set aspect ratio to 1
    min_val = df[["Predicted Value", "Actual Value"]].min().min()
    max_val = df[["Predicted Value", "Actual Value"]].max().max()
    line_length = max_val - min_val
    domain = (min_val - 0.1 * line_length, max_val + 0.1 * line_length)
    df_diagonal = pd.DataFrame({"Predicted Value": domain, "Actual Value": domain})

    diagonal_line = (
        alt.Chart(df_diagonal)
        .mark_line(strokeDash=[1, 1])
        .encode(
            x=alt.X("Predicted Value:Q", scale=alt.Scale(domain=domain)),
            y=alt.Y("Actual Value:Q", scale=alt.Scale(domain=domain)),
        )
    )

    points = (
        alt.Chart(df)
        .mark_line(point=True)
        .encode(
            x=alt.X("Predicted Value:Q", scale=alt.Scale(domain=domain)),
            y=alt.Y("Actual Value:Q", scale=alt.Scale(domain=domain)),
            tooltip=[
                alt.Tooltip("Predicted Value:Q", format=".2f"),
                alt.Tooltip("Actual Value:Q", format=".2f"),
            ],
        )
        .interactive()
    )

    return diagonal_line + points


def binary_tpr_fpr(y: np.array, preds: np.array) -> Tuple[np.array, np.array]:
    """Return the true postive rates and false postive rates for different binary classification thresholds.
    Inspired by: https://github.com/scikit-learn/scikit-learn/blob/95119c13af77c76e150b753485c662b7c52a41a2/sklearn/metrics/_ranking.py#L653
    (License: BSD-3)
    """
    # Sort by predicted probability
    sorted_prob_idx = np.argsort(preds)[::-1]

    y = y[sorted_prob_idx]
    preds = preds[sorted_prob_idx]

    # Choose unique probabilities
    distinct_threshold_idx = np.where(np.diff(preds))[0]
    threshold_idx = np.concatenate([distinct_threshold_idx, [y.size - 1]])

    # Sum the total true positives and false positives for each threshold
    tp_cumsum = np.cumsum(y)[threshold_idx]
    fp_cumsum = np.cumsum(1 - y)[threshold_idx]

    tp_cumsum = np.concatenate([[0], tp_cumsum])
    fp_cumsum = np.concatenate([[0], fp_cumsum])

    if fp_cumsum[-1] <= 0:
        logging.warning("No false values in y")
        fpr = np.repeat(0, fp_cumsum.shape)
    else:
        fpr = fp_cumsum / fp_cumsum[-1]

    if tp_cumsum[-1] <= 0:
        logging.warning("No true values in y")
        tpr = np.repeat(0, tp_cumsum.shape)
    else:
        tpr = tp_cumsum / tp_cumsum[-1]

    return tpr, fpr


def roc_curve(
    func: Callable[[pd.DataFrame], pd.Series],
    x: pd.DataFrame,
    y: pd.Series,
) -> alt.Chart:
    """Plot ROC curve for binary endpoints"""

    preds = func(x).to_numpy()
    y = y.to_numpy()

    tpr, fpr = binary_tpr_fpr(y, preds)

    roc_curve_df = pd.DataFrame({"False Positive Rate": fpr, "True Positive Rate": tpr})

    # Subset df for smaller vega specs
    subset = np.unique(np.round(np.linspace(0, len(roc_curve_df) - 1, num=1000)))
    roc_curve_df = roc_curve_df.iloc[subset, :]

    # Add the 45 degree line for the random classifier
    dotted_line_df = pd.DataFrame(
        {
            "False Positive Rate": [0, 1],
            "True Positive Rate": [0, 1],
        }
    )

    roc_curve = (
        alt.Chart(roc_curve_df)
        .mark_line()
        .encode(
            x=alt.X("False Positive Rate", axis=alt.Axis(format="%")),
            y=alt.Y("True Positive Rate", axis=alt.Axis(format="%")),
            tooltip=[
                alt.Tooltip("False Positive Rate", format=".2%"),
                alt.Tooltip("True Positive Rate", format=".2%"),
            ],
        )
        .interactive()
    )

    dotted_line = (
        alt.Chart(dotted_line_df)
        .mark_line(strokeDash=[1, 1])
        .encode(
            x=alt.X(
                "False Positive Rate",
                axis=alt.Axis(format="%"),
            ),
            y=alt.Y(
                "True Positive Rate",
                axis=alt.Axis(format="%"),
            ),
        )
    )

    return roc_curve + dotted_line
